@@ -1,16 +1,17 @@
-use std::alloc::{alloc, dealloc, Layout};
+use std::alloc::{alloc, dealloc, handle_alloc_error, Layout};
 use std::any::{type_name, TypeId};
 use std::cell::{Ref, RefCell, RefMut};
 use std::cmp::Reverse;
+use std::mem::align_of;
 use std::ops::{Deref, DerefMut};
-use std::ptr::{copy_nonoverlapping, NonNull};
+use std::ptr::copy_nonoverlapping;
 
 pub struct Archetype {
     types: Box<[TypeMetadata]>,
     layout: Layout,
     len: u32,
     cap: u32,
-    ptr: NonNull<u8>,
+    ptr: *mut u8,
 }
 
 impl Archetype {
@@ -25,7 +26,7 @@ impl Archetype {
             .unwrap_or(1);
 
         let layout = Layout::from_size_align(0, max_align).unwrap();
-        let ptr = NonNull::new(layout.align() as *mut u8).unwrap();
+        let ptr = max_align as *mut u8;
 
         Self {
             types: types.0.into(),
@@ -54,11 +55,9 @@ impl Archetype {
 
     #[must_use]
     pub unsafe fn free(&mut self, idx: u32, drop: bool) -> bool {
-        let ptr = self.ptr.as_ptr();
-
         if drop {
             for ty in &*self.types {
-                let ptr = ptr.add(ty.offset + ty.layout.size() * idx as usize);
+                let ptr = ty.base_pointer.add(ty.layout.size() * idx as usize);
 
                 (ty.drop)(ptr);
             }
@@ -70,8 +69,8 @@ impl Archetype {
             for ty in &*self.types {
                 let size = ty.layout.size();
 
-                let src_ptr = ptr.add(ty.offset + size * self.len as usize);
-                let dst_ptr = ptr.add(ty.offset + size * idx as usize);
+                let src_ptr = ty.base_pointer.add(size * self.len as usize);
+                let dst_ptr = ty.base_pointer.add(size * idx as usize);
 
                 copy_nonoverlapping(src_ptr, dst_ptr, size);
             }
@@ -86,14 +85,10 @@ impl Archetype {
         let len = self.len;
         self.len = 0;
 
-        let ptr = self.ptr.as_ptr();
-
         for ty in &*self.types {
             unsafe {
-                let ptr = ptr.add(ty.offset);
-
                 for idx in 0..len {
-                    let ptr = ptr.add(ty.layout.size() * idx as usize);
+                    let ptr = ty.base_pointer.add(ty.layout.size() * idx as usize);
 
                     (ty.drop)(ptr);
                 }
@@ -120,55 +115,49 @@ impl Archetype {
             .0
             .sort_unstable_by_key(|ty| Reverse(ty.layout.align()));
 
-        let mut max_align = 1;
-        let mut sum_size = 0;
         let mut new_offsets = Vec::<usize>::with_capacity(types.0.len());
+        let mut new_size = 0;
 
         for ty in types.0.iter() {
-            let size = ty.layout.size();
-            let align = ty.layout.align();
+            new_offsets.push(new_size);
 
-            max_align = max_align.max(align);
-
-            debug_assert!(sum_size % align == 0);
-            new_offsets.push(sum_size);
-
-            sum_size = sum_size
-                .checked_add(size.checked_mul(new_cap).unwrap())
+            new_size = new_size
+                .checked_add(ty.layout.size().checked_mul(new_cap).unwrap())
                 .unwrap();
         }
 
-        let new_layout = Layout::from_size_align(sum_size, max_align).unwrap();
-        let new_ptr = NonNull::new(if new_layout.size() != 0 {
+        let new_layout = Layout::from_size_align(new_size, self.layout.align()).unwrap();
+
+        let new_ptr = if new_layout.size() != 0 {
             unsafe { alloc(new_layout) }
         } else {
             new_layout.align() as *mut u8
-        })
-        .unwrap();
+        };
+        if new_ptr.is_null() {
+            handle_alloc_error(new_layout);
+        }
 
-        if self.layout.size() != 0 {
-            let old_ptr = self.ptr.as_ptr();
+        unsafe {
+            for (ty, new_offset) in types.0.iter_mut().zip(&new_offsets) {
+                let new_base_pointer = new_ptr.add(*new_offset);
 
-            unsafe {
-                for (ty, new_offset) in types.0.iter().zip(&new_offsets) {
-                    copy_nonoverlapping(
-                        old_ptr.add(ty.offset),
-                        new_ptr.as_ptr().add(*new_offset),
-                        ty.layout.size() * self.cap as usize,
-                    );
-                }
+                copy_nonoverlapping(
+                    ty.base_pointer,
+                    new_base_pointer,
+                    ty.layout.size() * self.cap as usize,
+                );
 
-                dealloc(old_ptr, self.layout);
+                ty.base_pointer = new_base_pointer;
+            }
+
+            if self.layout.size() != 0 {
+                dealloc(self.ptr, self.layout);
             }
         }
 
         self.layout = new_layout;
         self.ptr = new_ptr;
 
-        for (ty, new_offset) in types.0.iter_mut().zip(&new_offsets) {
-            ty.offset = *new_offset;
-        }
-
         self.cap = new_cap as u32;
     }
 }
@@ -177,9 +166,9 @@ impl Drop for Archetype {
     fn drop(&mut self) {
         self.clear();
 
-        if self.layout.size() != 0 {
-            unsafe {
-                dealloc(self.ptr.as_ptr(), self.layout);
+        unsafe {
+            if self.layout.size() != 0 {
+                dealloc(self.ptr, self.layout);
             }
         }
     }
@@ -227,9 +216,6 @@ impl Archetype {
 
         let mut dst_types = &*dst.types;
 
-        let src_ptr = src.ptr.as_ptr();
-        let dst_ptr = dst.ptr.as_ptr();
-
         for src_ty in &*src.types {
             let dst_ty = match dst_types.iter().position(|ty| ty.id == src_ty.id) {
                 Some(dst_ty) => {
@@ -241,8 +227,8 @@ impl Archetype {
 
             let size = src_ty.layout.size();
 
-            let src_ptr = src_ptr.add(src_ty.offset + size * src_idx as usize);
-            let dst_ptr = dst_ptr.add(dst_ty.offset + size * dst_idx as usize);
+            let src_ptr = src_ty.base_pointer.add(size * src_idx as usize);
+            let dst_ptr = dst_ty.base_pointer.add(size * dst_idx as usize);
 
             copy_nonoverlapping(src_ptr, dst_ptr, size);
         }
@@ -288,7 +274,7 @@ impl Archetype {
     {
         let ty = self.type_metadata::<C>(ty);
 
-        self.ptr.as_ptr().add(ty.offset).cast::<C>()
+        ty.base_pointer.cast::<C>()
     }
 }
 
@@ -362,7 +348,7 @@ struct TypeMetadata {
     id: TypeId,
     layout: Layout,
     drop: unsafe fn(*mut u8),
-    offset: usize,
+    base_pointer: *mut u8,
     borrow: RefCell<()>,
 }
 
@@ -372,7 +358,7 @@ impl Clone for TypeMetadata {
             id: self.id,
             layout: self.layout,
             drop: self.drop,
-            offset: 0,
+            base_pointer: self.layout.align() as *mut u8,
             borrow: Default::default(),
         }
     }
@@ -391,7 +377,7 @@ impl TypeMetadata {
             id: TypeId::of::<C>(),
             layout: Layout::new::<C>(),
             drop: drop_in_place::<C>,
-            offset: 0,
+            base_pointer: align_of::<C>() as *mut u8,
             borrow: Default::default(),
         }
     }
@@ -452,16 +438,22 @@ mod tests {
 
         assert_eq!(archetype.types.len(), 4);
 
-        let ty = archetype.find::<u64>().unwrap();
-        assert_eq!(archetype.types[ty as usize].offset, 0);
+        unsafe {
+            let ty = archetype.find::<u64>().unwrap();
+            let base_pointer = archetype.base_pointer::<u64>(ty);
+            assert_eq!(base_pointer, archetype.ptr.add(0).cast());
 
-        let ty = archetype.find::<u32>().unwrap();
-        assert_eq!(archetype.types[ty as usize].offset, 8 * 8);
+            let ty = archetype.find::<u32>().unwrap();
+            let base_pointer = archetype.base_pointer::<u32>(ty);
+            assert_eq!(base_pointer, archetype.ptr.add(8 * 8).cast());
 
-        let ty = archetype.find::<u16>().unwrap();
-        assert_eq!(archetype.types[ty as usize].offset, 8 * (8 + 4));
+            let ty = archetype.find::<u16>().unwrap();
+            let base_pointer = archetype.base_pointer::<u16>(ty);
+            assert_eq!(base_pointer, archetype.ptr.add(8 * (8 + 4)).cast());
 
-        let ty = archetype.find::<u8>().unwrap();
-        assert_eq!(archetype.types[ty as usize].offset, 8 * (8 + 4 + 2));
+            let ty = archetype.find::<u8>().unwrap();
+            let base_pointer = archetype.base_pointer::<u8>(ty);
+            assert_eq!(base_pointer, archetype.ptr.add(8 * (8 + 4 + 2)).cast());
+        }
     }
 }
