@@ -3,7 +3,6 @@ use std::collections::HashMap;
 use std::convert::TryInto;
 use std::hash::{BuildHasherDefault, Hasher};
 use std::num::NonZeroU32;
-use std::ops::BitXor;
 use std::sync::atomic::{AtomicU32, Ordering};
 
 use crate::archetype::{Archetype, Comp, CompMut, TypeMetadataSet};
@@ -14,9 +13,9 @@ pub struct World {
     pub(crate) entities: Vec<EntityMetadata>,
     free_list: Vec<u32>,
     pub(crate) archetypes: Vec<Archetype>,
-    insert_map: HashMap<(u32, TypeId), u32, BuildHasherDefault<IndexTypeIdHasher>>,
-    remove_map: HashMap<(u32, TypeId), u32, BuildHasherDefault<IndexTypeIdHasher>>,
-    transfer_map: HashMap<(u32, u32), u32, BuildHasherDefault<U32PairHasher>>,
+    insert_map: IndexTypeIdMap,
+    remove_map: IndexTypeIdMap,
+    transfer_map: U32PairMap,
 }
 
 impl Default for World {
@@ -174,19 +173,16 @@ impl World {
         let meta = &self.entities[ent.id as usize];
         assert_eq!(ent.gen, meta.gen, "Entity is stale");
 
-        let new_ty;
-
-        if let Some(ty) = self.insert_map.get(&(meta.ty, TypeId::of::<B>())) {
-            new_ty = *ty;
+        let new_ty = if let Some(ty) = self.insert_map.get(&(meta.ty, TypeId::of::<B>())) {
+            *ty
         } else {
-            let mut types = self.archetypes[meta.ty as usize].types();
-            B::insert(&mut types);
-
-            new_ty = Self::get_or_insert(&mut self.archetypes, types);
-
-            self.insert_map.insert((meta.ty, TypeId::of::<B>()), new_ty);
-            self.remove_map.insert((new_ty, TypeId::of::<B>()), meta.ty);
-        }
+            Self::insert_cold::<B>(
+                &mut self.archetypes,
+                &mut self.insert_map,
+                &mut self.remove_map,
+                meta.ty,
+            )
+        };
 
         let old_ty = meta.ty;
         let old_idx = meta.idx;
@@ -198,6 +194,28 @@ impl World {
         }
     }
 
+    #[cold]
+    #[inline(never)]
+    fn insert_cold<B>(
+        archetypes: &mut Vec<Archetype>,
+        insert_map: &mut IndexTypeIdMap,
+        remove_map: &mut IndexTypeIdMap,
+        old_ty: u32,
+    ) -> u32
+    where
+        B: Bundle,
+    {
+        let mut types = archetypes[old_ty as usize].types();
+        B::insert(&mut types);
+
+        let new_ty = Self::get_or_insert(archetypes, types);
+
+        insert_map.insert((old_ty, TypeId::of::<B>()), new_ty);
+        remove_map.insert((new_ty, TypeId::of::<B>()), old_ty);
+
+        new_ty
+    }
+
     /// Remove components for a given [Entity].
     ///
     /// # Example
@@ -219,19 +237,16 @@ impl World {
         let meta = &self.entities[ent.id as usize];
         assert_eq!(ent.gen, meta.gen, "Entity is stale");
 
-        let new_ty;
-
-        if let Some(ty) = self.remove_map.get(&(meta.ty, TypeId::of::<B>())) {
-            new_ty = *ty;
+        let new_ty = if let Some(ty) = self.remove_map.get(&(meta.ty, TypeId::of::<B>())) {
+            *ty
         } else {
-            let mut types = self.archetypes[meta.ty as usize].types();
-            B::remove(&mut types)?;
-
-            new_ty = Self::get_or_insert(&mut self.archetypes, types);
-
-            self.remove_map.insert((meta.ty, TypeId::of::<B>()), new_ty);
-            self.insert_map.insert((new_ty, TypeId::of::<B>()), meta.ty);
-        }
+            Self::remove_cold::<B>(
+                &mut self.archetypes,
+                &mut self.insert_map,
+                &mut self.remove_map,
+                meta.ty,
+            )?
+        };
 
         let old_ty = meta.ty;
         let old_idx = meta.idx;
@@ -245,6 +260,28 @@ impl World {
         }
     }
 
+    #[cold]
+    #[inline(never)]
+    fn remove_cold<B>(
+        archetypes: &mut Vec<Archetype>,
+        insert_map: &mut IndexTypeIdMap,
+        remove_map: &mut IndexTypeIdMap,
+        old_ty: u32,
+    ) -> Option<u32>
+    where
+        B: Bundle,
+    {
+        let mut types = archetypes[old_ty as usize].types();
+        B::remove(&mut types)?;
+
+        let new_ty = Self::get_or_insert(archetypes, types);
+
+        remove_map.insert((old_ty, TypeId::of::<B>()), new_ty);
+        insert_map.insert((new_ty, TypeId::of::<B>()), old_ty);
+
+        Some(new_ty)
+    }
+
     fn get_or_insert(archetypes: &mut Vec<Archetype>, types: TypeMetadataSet) -> u32 {
         let pos = archetypes
             .iter_mut()
@@ -313,16 +350,19 @@ impl World {
         self.free_list.push(ent.id);
 
         // get or insert new archetype in other
-        if let Some(ty) = self.transfer_map.get(&(meta.ty, other.tag)) {
-            new_meta.ty = *ty;
+        new_meta.ty = if let Some(ty) = self.transfer_map.get(&(meta.ty, other.tag)) {
+            *ty
         } else {
-            let types = self.archetypes[meta.ty as usize].types();
-
-            new_meta.ty = Self::get_or_insert(&mut other.archetypes, types);
-
-            self.transfer_map.insert((meta.ty, other.tag), new_meta.ty);
-            other.transfer_map.insert((new_meta.ty, self.tag), meta.ty);
-        }
+            Self::transfer_cold(
+                &self.archetypes,
+                &mut other.archetypes,
+                &mut self.transfer_map,
+                &mut other.transfer_map,
+                self.tag,
+                other.tag,
+                meta.ty,
+            )
+        };
 
         // move components from old to new archetype
         let old_archetype = &mut self.archetypes[meta.ty as usize];
@@ -350,6 +390,27 @@ impl World {
 
         ent
     }
+
+    #[cold]
+    #[inline(never)]
+    fn transfer_cold(
+        archetypes: &[Archetype],
+        other_archetypes: &mut Vec<Archetype>,
+        transfer_map: &mut U32PairMap,
+        other_transfer_map: &mut U32PairMap,
+        tag: u32,
+        other_tag: u32,
+        old_ty: u32,
+    ) -> u32 {
+        let types = archetypes[old_ty as usize].types();
+
+        let new_ty = Self::get_or_insert(other_archetypes, types);
+
+        transfer_map.insert((old_ty, other_tag), new_ty);
+        other_transfer_map.insert((new_ty, tag), old_ty);
+
+        new_ty
+    }
 }
 
 impl World {
@@ -534,6 +595,8 @@ macro_rules! impl_bundle_for_tuples {
 
 impl_bundle_for_tuples!(A, B, C, D, E, F, G, H, I, J);
 
+type IndexTypeIdMap = HashMap<(u32, TypeId), u32, BuildHasherDefault<IndexTypeIdHasher>>;
+
 #[derive(Default)]
 struct IndexTypeIdHasher(u64);
 
@@ -555,14 +618,14 @@ impl Hasher for IndexTypeIdHasher {
     }
 }
 
+type U32PairMap = HashMap<(u32, u32), u32, BuildHasherDefault<U32PairHasher>>;
+
 #[derive(Default)]
 struct U32PairHasher(u64);
 
 impl Hasher for U32PairHasher {
     fn write_u32(&mut self, val: u32) {
-        const K: u64 = 0x517cc1b727220a95;
-
-        self.0 = self.0.rotate_left(5).bitxor(val as u64).wrapping_mul(K);
+        self.0 = self.0 << 32 | val as u64;
     }
 
     fn write(&mut self, _val: &[u8]) {
@@ -570,7 +633,7 @@ impl Hasher for U32PairHasher {
     }
 
     fn finish(&self) -> u64 {
-        self.0
+        self.0.wrapping_mul(0x517cc1b727220a95)
     }
 }
 
@@ -876,4 +939,24 @@ mod tests {
         let comp = world2.get::<i32>(ent2).unwrap();
         assert_eq!(*comp, 23);
     }
+
+    #[test]
+    fn u32_pair_hasher_yields_uniformly_distributed_lower_bits() {
+        let mut histogram = [0; 128];
+
+        for i in 0..1024 {
+            for j in 0..128 {
+                let mut hasher = U32PairHasher::default();
+                hasher.write_u32(i);
+                hasher.write_u32(j);
+                let hash = hasher.finish();
+
+                histogram[hash as usize % histogram.len()] += 1;
+            }
+        }
+
+        for count in histogram {
+            assert_eq!(count, 1024 * 128 / histogram.len());
+        }
+    }
 }
